/* This code parses the output of dmalloc (log-trans),
 * matches alloc/free pairs, and prints out unfreed
 * blocksi (memory leaks).
 *
 * To use:
 * 	. install the dmalloc package
 * 	. link the code to be debugged using '-ldmalloc'
 * 	. put the following line in ~/.dmallocrc
 *		trace	log-non-free,log-trans,error-abort
 * 	. execute 'dmalloc -l logfile -i 100 trace'
 * 	. set the variable 'DMALLOC_OPTIONS' according to the output.
 * 	. run the code
 * 	. type 'dmalloc-check logfile' to list memoey leaks
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct ENTRY {
    void *ptr;
    int  size;
    char loc[128];
    struct ENTRY *next;
} entry;

entry *root=NULL;

void add_entry(void *ptr, int size, char *loc)
{
    entry *p;

    if ( (p=malloc(sizeof(entry))) == NULL )
    {
	fprintf(stderr, "Memory allocation error\n");
	exit(1);
    }
    p->ptr=ptr;
    p->size=size;
    strncpy(p->loc, loc, sizeof(p->loc)-1);
    p->loc[sizeof(p->loc)-1]='\0';
    p->next=NULL;

    if ( root == NULL ) {
	root = p;
    } else {
	entry *p0=root;
	while (p0->next!=NULL) p0=p0->next;
	p0->next=p;
    }
}

void del_entry(void *ptr, int size, char *loc)
{
    entry *p=root, *p0=NULL;

    (void)loc;
    
    while ( p!=NULL ) {
	if ( p->ptr == ptr ) break;
	p0 = p;
	p = p->next;
    }

    if (p==NULL) return;	// not found
    if ( p->size != size ) fprintf(stderr, "warning: size mismatch.\n");

    if ( p==root ) root=p->next; else p0->next=p->next;

    free(p);
}

void *get_ptr(char *str)
{
    unsigned ptr;
    *(strchr(str, '|'))='\0';
    sscanf(str+1, "%x", &ptr);
    return (void *) ptr;
}

int main(int argc, char **argv)
{
    int linecnt=0;
    FILE *f;
    void *ptr;
    int size;
    char line[1024];
    char s1[128], s2[128], s3[128];

    if (argc!=2) {
	fprintf(stderr, "Usage: %s logfile\n", argv[0]);
	exit(3);
    }

    if ( (f=fopen(argv[1], "rt")) == NULL ) {
	fprintf(stderr, "Cannot open \"%s\"\n", argv[1]);
	exit(2);
    }

    		/* For miniChinput */
    		while ( !feof(f) ) {
		    if (fgets(line, sizeof(line), f)==NULL) break;
		    ++linecnt;
		    if (sscanf(line, "%*s %*s %s %s", s1, s2)!=2) continue;
		    if (!strcmp(s1, "ending") && !strcmp(s2, "time")) break;
		}

    while (!feof(f)) {
	if (fgets(line, sizeof(line), f)==NULL) break;
	if ( ! (++linecnt%100) ) fprintf(stderr, "Reading line %d\r", linecnt);
	if (sscanf(line, "%*s %*s %*s %s", s1)!=1) continue;

	if (!strcmp(s1, "alloc:")) {
	    if (sscanf(line, "%*s %*s %*s %*s %*s %s %*s %s %*s %*s %s",
			s1, s2, s3)!=3) continue;
	    size=atoi(s2);
	    ptr=get_ptr(s3);
	    add_entry(ptr, size, s1);
	}
	
	if (!strcmp(s1, "free:")) {
	    if (sscanf(line, "%*s %*s %*s %*s %*s %s %*s %s %*s %s",
			s1, s3, s2)!=3) continue;
	    size=atoi(s2);
	    ptr=get_ptr(s3);
	    del_entry(ptr, size, s1);
	}
    }

    fclose(f);

    /* Dump output */
    fprintf(stderr, "Unfreed memory blocks:\n");
    while (root!=NULL) {
	printf("\tptr=%p, %d, %s\n", root->ptr, root->size, root->loc);
	root=root->next;
    }

    return 0;
}
